#  tests for joblib-0.12.3-py35_0 (this is a generated file);
print('===== testing package: joblib-0.12.3-py35_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from joblib import Memory, Parallel, delayed
#  --- run_test.py (end) ---

print('===== joblib-0.12.3-py35_0 OK =====');
print("import: 'joblib'")
import joblib

